/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.anfrage;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;

import com.bitctrl.Constants;

import de.bsvrz.dav.daf.main.ClientSenderInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SendSubscriptionNotConfirmed;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.puk.param.lib.ParametrierungsAnfrageTyp;
import de.bsvrz.puk.param.param.ParamMessage;
import de.bsvrz.puk.param.param.Parametrierung;
import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Abstrakte Basisklasse zur Repräsentation eines Beantworters auf eine Anfrage,
 * die die Parametrierung über die Parametrierungsschnittstelle erhalten hat
 * (atg.parameterSchnittstelle). Für eine Anfrage wird ein Thread gestartet, der
 * die gesuchten Informationen ermittelt und diese als Sender an die anfragende
 * Applikation übermittelt.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
public abstract class ParameterAnfrageBeantworter extends Thread implements ClientSenderInterface {

	/**
	 * Exception, wenn eine Antwort nicht an den Auftraggeber versendet werden
	 * konnte.
	 */
	private class AntwortException extends Exception {

		/** Versions-ID zur Serialisierung. */
		private static final long serialVersionUID = 1L;

		/**
		 * Konstruktor.
		 *
		 * @param msg
		 *            der Meldungstext
		 * @param exception
		 *            optionale Exception
		 */
		AntwortException(final String msg, final Throwable exception) {
			super(msg, exception);
		}

		/**
		 * Konstruktor.
		 *
		 * @param msg
		 *            der Meldungstext
		 */
		AntwortException(final String msg) {
			super(msg);
		}
	}

	/** Logger für Debug-Meldungen. */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * die Datenbeschreibung zum Beantworten von Anfragen über die Schnittstelle
	 * "atg.parameterierungsAntwort".
	 */
	private final DataDescription antwortDesc;

	/**
	 * Factory zum Erstellen eines konkreten Anfragebeantworters in Abhängigkeit
	 * vom empfangenen Anfragetyp.<br>
	 * Unterstützt werden Anfragen mit folgenden Typen:
	 * <ol>
	 * <li>{@link KonfigurierbareObjekteErmittler}: zum Ermitteln der Objekte,
	 * die durch Setzen des übergebenen Parameterdatensatzes potentiell mit
	 * Parametern versorgt werden. Das ist nur sinnvoll für Objekte, die eine
	 * Hierarchie beschreiben oder Typ-Objekte.</li>
	 * <li>{@link ParameterLesenErmittler}: zum Ermitteln der konkreten
	 * Parameter für die übergebene Spezifikation. über diese Anfrage können
	 * auch die Parameter ermittelt werden, die an Typ-Objekten bzw.
	 * Hierarchieobjekten für die versorgten untergeordneten Objekte gelten.
	 * </li>
	 * <li>{@link ParameterSchreibenErmittler}: zum Schreiben der Parameter für
	 * die übergebene Spezifikation.</li>
	 * <li>{@link SucheWurzelErmittler}: zum Ermitteln der Basisobjektes, aus
	 * dem die Parameter für eine übergebene Parameterspezifikation stammen.
	 * </li>
	 * </ol>
	 *
	 * Wird ein nicht unterstützter Anfragetyp übermittelt, wirft die Funktion
	 * eine {@link IllegalArgumentException}.
	 *
	 * @param parametrierung
	 *            die Parametrierung, die die Anfrage erhalten hat
	 * @param daten
	 *            die Daten die mit der Anfrage empfangen wurden
	 * @return der erzeugte Beantworter
	 */
	public static ParameterAnfrageBeantworter erzeugeBeantworter(final Parametrierung parametrierung,
			final Data daten) {

		final SystemObject anfrageObjekt = daten.getReferenceValue("Absender").getSystemObject();
		final int anfrageIndex = daten.getUnscaledValue("AnfrageIndex").intValue();
		final ParametrierungsAnfrageTyp typ = ParametrierungsAnfrageTyp
				.getTyp(daten.getUnscaledValue("AnfrageTyp").intValue());
		final byte[] bytes = daten.getArray("Daten").asUnscaledArray().getByteArray();
		final ByteArrayInputStream input = new ByteArrayInputStream(bytes);
		Object datenObjekt = null;
		try {
			final ObjectInputStream oi = new ObjectInputStream(input);
			datenObjekt = oi.readObject();
		} catch (final IOException e) {
			// die Daten werden als null-Referenz übergeben. Der Fehler wird mit
			// der Antwort generiert.
			LogTools.log(ParameterAnfrageBeantworter.LOGGER, ParamMessage.ErrorLesenAnfrageDaten, e.getMessage());
		} catch (final ClassNotFoundException e) {
			// die Daten werden als null-Referenz übergeben. Der Fehler wird mit
			// der Antwort generiert.
			LogTools.log(ParameterAnfrageBeantworter.LOGGER, ParamMessage.ErrorLesenAnfrageDaten, e.getMessage());
		}

		ParameterAnfrageBeantworter beantworter = null;
		switch (typ) {
		case SUCHE_AENDERBARE_OBJEKTE:
			beantworter = new KonfigurierbareObjekteErmittler(parametrierung, anfrageObjekt, anfrageIndex, datenObjekt);
			break;
		case SUCHE_PARAMETER_WURZEL:
			beantworter = new SucheWurzelErmittler(parametrierung, anfrageObjekt, anfrageIndex, datenObjekt);
			break;
		case PARAMETER_LESEN:
			beantworter = new ParameterLesenErmittler(parametrierung, anfrageObjekt, anfrageIndex, datenObjekt);
			break;
		case PARAMETER_SCHREIBEN:
			beantworter = new ParameterSchreibenErmittler(parametrierung, anfrageObjekt, anfrageIndex, datenObjekt);
			break;
		case PARAMETER:
			beantworter = new ControlAnfrageBeantworter(parametrierung, anfrageObjekt, anfrageIndex);
			break;
		default:
			throw new IllegalArgumentException("Unerwarteter Anfragetyp: " + typ);
		}

		return beantworter;
	}

	/**
	 * das anfragende Objekt, üblicherweise ein Objekt vom Typ "Applikation".
	 */
	private final SystemObject anfrager;

	/**
	 * die Parametrierung, die die Anfrage empfangen hat.
	 */
	private final Parametrierung parametrierung;

	/**
	 * der Index der Anfrage, wird mit der Antwort zurückgesendet.
	 */
	private final int index;

	/**
	 * markiert die Rückmeldung der Sendesteuerung zum Versenden einer Antwort.
	 */
	private boolean sendeBereitschaft;

	/** Mutex zur Synchronisation der Sendebereitschaft. */
	private final Object sendeBereitschaftMutex = new Object();

	/**
	 * Konstruktor.<br>
	 * Die Funktion erzeugt einen Anfragebeantworter der eine Anfrage mit Hilfe
	 * des übergebenen Parametrierungsmodul bearbeitet. Das übergebene
	 * Anfrageobjekt muss ein Objekt vom Typ "typ.applikation" sein, da über
	 * dessen Attributgruppe "atg.paremetrierungSchnittstelle" das Ergebnis
	 * übermittelt wird. Der Index wird im Ergebnis zurückgeliefert.
	 *
	 * @param parametrierung
	 *            die Parametrierung, die die Anfrage erhalten hat.
	 * @param anfrager
	 *            das anfragende Applikationsobjekt
	 * @param index
	 *            der Index der Anfrage
	 */
	protected ParameterAnfrageBeantworter(final Parametrierung parametrierung, final SystemObject anfrager,
			final int index) {

		this.antwortDesc = parametrierung.getAntwortDesc();
		this.parametrierung = parametrierung;
		this.anfrager = anfrager;
		this.index = index;
	}

	@Override
	public void dataRequest(final SystemObject object, final DataDescription dataDescription, final byte state) {
		synchronized (sendeBereitschaftMutex) {
			if (state == ClientSenderInterface.START_SENDING) {
				sendeBereitschaft = true;
			} else {
				sendeBereitschaft = false;
			}
			synchronized (this) {
				this.notifyAll();
			}
		}
	}

	/**
	 * die Funktion ermittelt aus den übergebenen Daten eine Antwort, die später
	 * an den Anfrager als Attribut im Datensatz
	 * "atg.parametrierungSchnittstelle" zurückgesendet wird.
	 *
	 * @return die Antwort als serialisierbares Objekt
	 */
	protected abstract Serializable ermittleAntwort();

	/**
	 * liefert das Systemobjekt, das die anfragende Applikation in der
	 * Datenverteilerkonfiguration repräsentiert.
	 *
	 * @return den Anfrager
	 */
	SystemObject getAnfrager() {
		return anfrager;
	}

	/**
	 * liefert den Typ der Antwort auf die entsprechende Anfrage. Der Typ der
	 * Antwort hängt vom Typ der Anfrage ab, ist aber durch den gleichen Zustand
	 * definiert, da Anfrage und Antwort mit der gleichen Attributgruppe
	 * versendet werden.
	 *
	 * @return den Antworttyp
	 */
	protected abstract ParametrierungsAnfrageTyp getAntwortTyp();

	/**
	 * liefert den Index der Anfrage.
	 *
	 * @return den Index
	 */
	protected int getIndex() {
		return index;
	}

	/**
	 * liefert die Parametrierung, die die Anfrage erhalten hat.
	 *
	 * @return parametrierung die Parametrierung
	 */
	protected Parametrierung getParametrierung() {
		return parametrierung;
	}

	@Override
	public boolean isRequestSupported(final SystemObject object, final DataDescription dataDescription) {
		return true;
	}

	@Override
	public void run() {
		try {
			versendeAntwort();
		} catch (final AntwortException e) {
			Debug.getLogger().error(e.getLocalizedMessage(), e);
		}
	}

	/**
	 * die Funktion versendet die Antwort an den Anfrager.<br>
	 * Dazu meldet sie sich als Sender für das bei der Initialisierung der
	 * Klasse ermittelte Applikationsobjekt und die Attributgruppe
	 * "atg.parmetrierungSchnittstelle" an, füllt den Datensatz und versendet
	 * diesen nach Empfang der Sendebestätigung. Nach dem Versand der Daten
	 * meldet sich die Parametrierung als Sender ab.
	 *
	 * @throws AntwortException
	 *             die Antwort konnte nicht erfolgreich versendet werden
	 */
	private void versendeAntwort() throws AntwortException {

		try {
			parametrierung.getVerbindung().subscribeSender(this, anfrager, antwortDesc, SenderRole.sender());
		} catch (final OneSubscriptionPerSendData e) {
			LogTools.log(ParameterAnfrageBeantworter.LOGGER, ParamMessage.ErrorSendeAnmeldungAntwort, e.getMessage());
			return;
		}

		final Data data = parametrierung.getVerbindung().createData(antwortDesc.getAttributeGroup());
		data.getReferenceValue("Absender").setSystemObject(parametrierung.getVerbindung().getLocalApplicationObject());
		data.getUnscaledValue("AnfrageIndex").set(index);
		data.getUnscaledValue("AnfrageTyp").set(getAntwortTyp().getCode());
		final Serializable antwort = ermittleAntwort();
		try {
			final ByteArrayOutputStream byteStream = new ByteArrayOutputStream();
			final ObjectOutputStream os = new ObjectOutputStream(byteStream);
			os.writeObject(antwort);
			final byte[] bytes = byteStream.toByteArray();

			final Data.NumberArray array = data.getUnscaledArray("Daten");
			array.setLength(bytes.length);
			for (int idx = 0; idx < bytes.length; idx++) {
				array.getValue(idx).set(bytes[idx]);
			}
		} catch (final IOException e) {
			throw new RuntimeException(e);
		}

		try {
			synchronized (this) {
				final long start = System.currentTimeMillis();
				while (!sendeBereitschaft) {
					try {
						this.wait(Constants.MILLIS_PER_MINUTE);
					} catch (final InterruptedException e) {
						throw new RuntimeException("Anfragebeantworter wurde unerwartet unterbrochen", e);
					}
					if (System.currentTimeMillis() - start > 2 * Constants.MILLIS_PER_MINUTE) {
						parametrierung.getVerbindung().unsubscribeSender(this, anfrager, antwortDesc);
						throw new AntwortException(
								"Anmeldung als Sender für die Beantwortung einer Antwort wurde nicht innerhalb von 120 Sekunden bestätigt");
					}
				}
			}
			parametrierung.getVerbindung()
					.sendData(new ResultData(anfrager, antwortDesc, System.currentTimeMillis(), data));
		} catch (final SendSubscriptionNotConfirmed e) {
			parametrierung.getVerbindung().unsubscribeSender(this, anfrager, antwortDesc);
			throw new AntwortException(
					"Anmeldung als Sender für die Beantwortung einer Antwort wurde nicht innerhalb von 60 Sekunden bestätigt",
					e);
		}

		parametrierung.getVerbindung().unsubscribeSender(this, anfrager, antwortDesc);
	}
}
