/*
 * Segment 8 Parametrierung und Konfiguration, SWE 8.2 Parametrierung
 * Copyright (C) 2007-2015 BitCtrl Systems GmbH
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 51
 * Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.puk.param.param;

import java.util.HashSet;
import java.util.Set;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.puk.param.lib.ParameterInfo;
import de.bsvrz.sys.funclib.bitctrl.daf.DaVKonstanten;
import de.bsvrz.sys.funclib.debug.Debug;

/**
 * Klasse zum Einlesen der Standardparameter, die an einem Objekt mit dem Aspekt
 * "parameterDefault" abgelegt sind.<br>
 * Diese Art der Festlegung von Standardparametern ist aus
 * Kompatibilitätsgründen enthalten und muss durch einen entsprechenden
 * Parameter bein Starten der Parametrierung aktiviert werden. Die
 * Standardparameter werden normalerweise durch die Attributgruppe
 * "atg.parameterDatensätze" beschrieben.
 *
 * @author BitCtrl Systems GmbH, Peuker
 */
class DefaultParameterReader {

	/**
	 * Eintrag mit dem registriert wird, ob die Standardparameter für eine
	 * Typ-Attributgruppen-Kombination bereits als Block ausgelesen wurden.
	 */
	private static class DefaultCacheItem {

		/** die ID der Attributgruppe. */
		private final long atgid;

		/** die ID des Typs. */
		private final long typid;

		/**
		 * Konstruktor, erzeugt einen Eintrag für die Speicherung der
		 * Standardparameter.
		 *
		 * @param typ
		 *            der Typ
		 * @param atg
		 *            die Attributgruppe
		 */
		DefaultCacheItem(final SystemObjectType typ, final AttributeGroup atg) {
			this.typid = typ.getId();
			this.atgid = atg.getId();
		}

		@Override
		public int hashCode() {
			final int prime = 31;
			int result = 1;
			result = prime * result + (int) (atgid ^ atgid >>> 32);
			result = prime * result + (int) (typid ^ typid >>> 32);
			return result;
		}

		@Override
		public boolean equals(final Object obj) {
			if (this == obj) {
				return true;
			}
			if (obj == null) {
				return false;
			}
			if (getClass() != obj.getClass()) {
				return false;
			}
			final DefaultCacheItem other = (DefaultCacheItem) obj;
			if (atgid != other.atgid) {
				return false;
			}
			if (typid != other.typid) {
				return false;
			}
			return true;
		}
	}

	/** Logger für Debugausgaben. */
	private static final Debug LOGGER = Debug.getLogger();

	/**
	 * der Aspekt der konfigurierten Standardparameter eines Objekts (soll
	 * demnächst entfallen).
	 */
	private final Aspect defaultAsp;

	/**
	 * Menge der Typ-Attributgruppe-Kombinationen, für die die Daten bereits
	 * blockweise ausgelesen wurden.
	 */
	private final Set<DefaultCacheItem> defaultCache = new HashSet<>();

	/** die verwendete Datenverteilerverbindung. */
	private final ClientDavInterface verbindung;

	/**
	 * Konstruktor, erzeugt eine Instanz des Readers für Standardwerte.
	 *
	 * @param parametrierung
	 *            das verwendete Parametrierungsmodul
	 */
	DefaultParameterReader(final Parametrierung parametrierung) {
		this.verbindung = parametrierung.getVerbindung();
		defaultAsp = parametrierung.getVerbindung().getDataModel().getAspect(DaVKonstanten.ASP_PARAMETER_DEFAULT);
	}

	/**
	 * liefert die Standarddaten für die übergebene Parameterspezifikation.
	 *
	 * @param info
	 *            die Spezifikation
	 * @return die Daten oder <i>null</i>, wenn keine konfiguriert sind
	 */
	Data getDaten(final ParameterInfo info) {
		final DefaultCacheItem item = new DefaultCacheItem(info.getTyp(), info.getAtg());
		if (!defaultCache.contains(item)) {
			verbindung.getDataModel().getConfigurationData(info.getTyp().getElements(), info.getAtg(), defaultAsp);
			defaultCache.add(item);
		}

		final Data daten = info.getObjekt().getConfigurationData(info.getAtg(), defaultAsp);

		if (daten != null) {
			DefaultParameterReader.LOGGER.finer("Default-Aspekt-Daten ermittelt:", info);
			DefaultParameterReader.LOGGER.finest("Daten:", daten);
		} else {
			DefaultParameterReader.LOGGER.finest("Keine Default-Aspekt-Daten ermittelt:", info);
		}
		return daten;
	}
}
